//<c>   Copyright 1990-1999 by Gerry J. Danen; all rights reserved.

//<t>   Convert Ascii text to PostScript
//
//      TO DO:
//              /pgxxx-yyy = print from page xxx to yyy (if xxx missing, start
//              at page 1, if yyy missing, go to end of file)

#include"gd_tools.h"
#include"gd_extrn.h"
#include<stdarg.h>


#define pgVERS_ "V3.40"
#define pgTITL_ "Print ASCII Files on a PostScript Printer"
#define pgCOPR_ "1990"


#if     (! LARGE_MODEL)
  #error LARGE MEMORY MODEL REQUIRED
#endif


#define WIDTH           8.5
#define HEIGHT          11.0
#define DIR_SEP         '\\'


#define PORTRAIT_HDR    0.29
#define LANDSCAPE_HDR   0.22
#define PIXELS_INCH     72
#define MAXFILENAME     20


double  margin = 1.2 ;          // Total margin (top+bottom) or (left+right)
int     left_margin = 0 ;       // Left margin in characters
int     column = 0 ;            // Column number (in current line)
int     line = 0 ;              // Line number (in current page)
int     line_number = 0 ;       // Source line number
int     first_page = 0 ;        // First page for a file
int     nonprinting_chars = 0 ; // Number of nonprinting chars
int     chars = 0 ;             // Number of total chars
int     prefix_width = 0 ;      // Width in characters for line prefix
int     pages = 0 ;             // Number of logical pages printed
int     sheets = 0 ;            // Number of physical pages printed
int     linesperpage = 0 ;      // Lines per page
int     columnsperline = 0 ;    // Characters per output line

double  font_size = 0.0;        // Size of a char for body font
int     numbering = FALSE;      // Line numbering option
int     folding = TRUE;         // Line folding option
int     restart = TRUE;         // Restart page number at each file option
int     only_printable = FALSE; // Replace non printable char by space option
int     interpret = TRUE;       // Interpret TAB, FF and BS chars option
int     print_binaries = FALSE; // Force printing for binary files
int     copies_number = 1;      // Number of copies to print
int     landscape = TRUE;       // Otherwise portrait format sheets
int     wide_pages = FALSE;     // TRUE implies landscape, not twinpage
int     twinpage = TRUE;        // 2 pages per sheet if true, 1 otherwise
int     no_border = FALSE;      // TRUE if user don't want the border
int     no_header = FALSE;      // TRUE if user don't want the header
BOOL    reps_132 = FALSE;       // TRUE if user wants 132-column x 66-line reports
BOOL    quiet_mode = FALSE ;    // TRUE if user wants no messages, unless there are problems
int     column_width = 8;       // default column tab width (8)
UCHAR   ps_header[ FS_LEN ] ;   // PostScript header file


void    cut_filename(char *old_name,char *new_name);
int     cut_line(void);
int     fold_line(char *name);
int     is_binaryfile(char *name);
int     printchar(unsigned char c);
void    print_file(char *name);
void    print_header(void);
void    program_init(char *pg_name);
void    skip_page(void);
void    usage(void);


void    abort_job( STRING msg, ... )
{
    UCHAR       mbuf[161];
    va_list     arg_ptr ;

    logon(1);
    va_start( arg_ptr, msg );
    vsprintf( mbuf, msg, arg_ptr );
    va_end( arg_ptr );
    fprintf( stderr, "\nProgram %s: fatal error!\a\n%s\n\nPress [Enter]...", pgm_name, mbuf );
    getchar();
    exit( 99 );
}


void    usage( void )
{
    logon(1);
    fprintf( stderr, "\nUsage:\t%s [options] [file_1 file_2 ... file_n]\n\n", pgm_name );
    fprintf( stderr,
                "    options:\t/?\n\t\t  or\n\t\t/h\tprint this information\n"
                "\t\t/#num\tnumber of copies to print\n"
                "\t\t/fnum\tfont size, num is a float number\n"
                "\t\t/[n]f\t[don't] fold lines that are too large\n"
                "\t\t/[n]i\t[don't] interpret special chars like tab, bs and ff\n"
                "\t\t/[n]n\t[don't] number output lines\n"
                "\t\t/[n]p\t[don't] print in portrait mode\n"
                "\t\t/[n]r\t[don't] restart page number @ 1 for each file\n"
                "\t\t/[n]v\t[don't] replace non-printing chars with spaces\n"
                "\t\t/[n]w\t[don't] print in wide format\n" );
    fprintf( stderr, "hit a key..." );
    getch();
    fprintf( stderr, "\r            \r" );
    fprintf( stderr,
                "\t\t/l\tleft margin in characters\n"
                "\t\t/nb\tdon't force printing binary files\n"
                "\t\t/nh\tdon't print the header\n"
                "\t\t/ns\tdon't print surrounding borders\n"
                "\t\t/q \tquiet mode \n"
                "\t\t/tn\tset tab size to n\n"
                "\t\t/132\tprint in 132x66 format\n"
                "\t\t/@fname\tuse fname as PostScript init file\n" );
    exit( BAD_EXIT );
}


void    print_file( char *name )
{
    register int c ;
    int start_line, continue_exit, char_width, start_page ;
    char new_name[MAXFILENAME+1];
    char *p;

    print_header();                     //  PostScript prologue

    //  Boolean to indicates that previous char is \n (or interpreted \f)
    //  and a new page would be started, if more text follows

    start_page = FALSE;

    //  Printing binary files is not very useful. We stop printing
    //  if we detect one of these files. Our heuristic to detect them:
    //  if 50% characters of first page are non-printing characters,
    //  the file is a binary file.
    //  Option -b force binary files impression.

    first_page = TRUE;
    nonprinting_chars = chars = 0;

    //  Preprocessing (before printing):
    //  - TABs expansion (see interpret option)
    //  - FF and BS interpretation
    //  - replace non printable characters by a space or a char sequence like:
    //      ^X for ascii codes < 0x20 (X = [@, A, B, ...])
    //      ^? for del char
    //      M-c for ascii codes > 0x3f
    //  - prefix parents and backslash ['(', ')', '\'] by backslash
    //    (escape character in PostScript)

    column = 0;
    line = line_number = 0;
    start_line = TRUE;

    if ( strlen(name) > MAXFILENAME )
    {
        cut_filename(name, new_name);
        name = new_name;
    }
    putchar( '(' );
    for ( p = name ; *p != EOS ; )
        printchar( *p++ );
    printf( ") newfile\n" );

    if ( restart )
        printf( "/sheet 1 def\n" );

    pages = 0;
    skip_page();

    c = getchar();
    while ( c != EOF )
    {
        //  meta-character
        if ( c > 0177 )
        {
            c = '~' ;
        }
        //  form feed
        if ( c == '\f' && interpret )
        {
            //  close current line
            if ( ! start_line )
            {
                printf( ") s\n" );
                start_line = TRUE;
            }

            //  start a new page?
            if ( start_page )
                skip_page();

            //  close current page and begin another
            printf( "endpage\n" ) ;
            start_page = TRUE;

            //  verification for binary files
            if ( first_page && is_binaryfile(name) )
                return;
            line = 0;
            if ( (c = getchar()) == EOF )
                break;
        }

        //  start a new line?
        if ( start_line )
        {
            if ( start_page )
            {                   //  only if there is something to print!
                skip_page();
                start_page = FALSE ;
            }
            if ( numbering )
                printf( "(%*s%-5d ", prefix_width-6, " ", ++line_number );
            else
                printf( "(%*s", prefix_width, " " );
            start_line = FALSE;
        }

        //  interpret each character
        switch ( c )
        {
            case '\b' :
                if ( ! interpret )
                    goto print;
                if ( column )
                    column--;
                putchar(c);
                break;
            case '\n' :
                column = 0;
                start_line = TRUE;
                printf( ") s\n" );
                if ( ++line >= linesperpage )
                {
                    printf( "endpage\n" );
                    start_page = TRUE ;
                    if ( first_page && is_binaryfile(name) )
                        return;
                    line = 0;
                }
                break;
            case '\t' :
                if ( interpret )
                {
                    continue_exit = FALSE;
                    do  {
                        if ( ++column + prefix_width > columnsperline )
                        {
                            if ( folding )
                            {
                                if ( fold_line(name) == FALSE )
                                    return;
                            }
                            else
                            {
                                c = cut_line();
                                continue_exit = TRUE;
                                break;
                            }
                        }
                        putchar(' ');
                    }
                    while ( column % column_width );
                    if ( continue_exit )
                        continue;
                    break;
                }
            default :
print:
                if ( only_printable )
                {
                    char_width = 1;
                }
                else
                {
                    char_width = c > 0177 ? 2 : 0 ;
                    char_width += c < ' ' || c == 0177 ? 2 : 1 ;
                }
                if ( prefix_width + (column += char_width) > columnsperline )
                {
                    if ( folding )
                    {
                        if ( fold_line(name) == FALSE )
                            return;
                    }
                    else
                    {
                        c = cut_line();
                        continue;
                    }
                }
                nonprinting_chars += printchar( (UCHAR) c );
                chars += 1 ;
                break;
        }
        c = getchar();
    }

    if ( ! start_line )
        printf( ") s\n" );
    if ( ! start_page )
        printf( "endpage\n" );
}


//      Cut long filenames

void    cut_filename( char *old_name, char *new_name )
{
    register char *p;
    register int i;

    p = old_name + (strlen(old_name)-1);
    while ( p >= old_name && *p != DIR_SEP )
        p--;

    for ( i = 0, p++; *p != EOS && i < MAXFILENAME; i++ )
        *new_name++ = *p++ ;
    *new_name = EOS ;
}


//      Fold a line too long

int     fold_line( char *name )
{
    column = 0;
    printf( ") s\n" );
    if ( ++line >= linesperpage )
    {
        printf( "endpage\n" );
        skip_page();
        if ( first_page && is_binaryfile(name) )
            return( FALSE );
        line = 0;
    }
    printf( "(%*s", prefix_width, " " );
    return( TRUE );
}


//      Cut a text line too long to the size of a page line

int     cut_line( void )
{
    int c;

    while ( (c = getchar()) != EOF && c != '\n' && c != '\f' )
        ;
    return( c );
}


//      Print a char in a form acceptable by PostScript printers

int     printchar( UCHAR c )
{
    if ( c >= ' ' && c < 0177 )
    {
        if ( c == '(' || c == ')' || c == '\\' )
            putchar('\\');
        putchar(c);
        return 0;
    }

    if ( only_printable )
    {
        putchar(' ');
        return 1;
    }

    if ( c > 0177 )
    {
        printf( "M-" );
        c &= 0177;
    }
    if ( c < ' ' )
    {
        putchar( '^' );
        if ( (c = c + (UCHAR) '@') == (UCHAR) '(' || c == (UCHAR) ')' || c == (UCHAR) '\\' )
            putchar( '\\' );
        putchar( c );
    }
    else
    if ( c == 0177 )
        printf( "^?" );
    else
        putchar(c);

    return( 1 );
}


//      Begin a new physical page

void    skip_page( void )
{
    pages += 1 ;
    if ( twinpage == FALSE || (pages & 0x1) )
    {
        sheets += 1 ;
        printf( "%%%%Page: %d %d\n", sheets, sheets );
    }
    printf( "startpage\n" );
}


//      See if we have a binary file

BOOL    is_binaryfile( char *name )
{
    first_page = FALSE;
    if ( ! print_binaries && (nonprinting_chars*100 / chars) >= 75 )
    {
        fprintf( stderr, "%s is a binary file: printing aborted\n", name );
        return TRUE;
    }
    return FALSE;
}


void    print_header( void )
{
    register int c;
    FILE *f;
    static      BOOL already_done = FALSE ;

    if ( already_done )
        return ;
    already_done = TRUE ;
    if ( (f = fopen(ps_header, "r")) == NULPTR )
        abort_job( "PostScript init file %s missing.", ps_header );
    if ( ! quiet_mode )
        fprintf( stderr, "Initializing printer from %s\n", ps_header );

    //  Initialize some PostScript variables
    printf( "%%!PS-Adobe\n%%---------\n" );
    printf( "%% Program\t: %s\n", pgm_name );
    printf( "%% Version\t: %s\n", pgm_vers );
    printf( "%% Title\t\t: %s\n", pgm_title );
    printf( "%% Copyright\t: %s\n", pgm_copyrite );
//  printf( "%% Copyright\t: %s - %s\n", pgm_copyrite, pgm_rights );
    printf( "%% CreationDate\t: %s %s\n\n", TodaysDate(), TodaysTime() );

    printf( "/$%sdict 100 dict def\n", pgm_name );
    printf( "$%sdict begin\n", pgm_name );
    printf( "%% Initialize page description variables.\n" );
    printf( "/inch {%d mul} bind def\n", PIXELS_INCH );
    printf( "/landscape %s def\n", landscape ? "true" : "false" );
    printf( "/twinpage %s def\n", twinpage ? "true" : "false" );
    printf( "/sheetheight %g inch def\n", HEIGHT );
    printf( "/sheetwidth %g inch def\n", WIDTH );
    printf( "/margin %g inch def\n", margin );
    printf( "/noborder %s def\n", no_border ? "true" : "false" );
    if ( no_header )
    {
        printf( "/noheader true def\n" );
        printf( "/headersize 0.0 def\n" );
    }
    else
    {
        printf( "/noheader false def\n" );
        printf( "/headersize %g inch def\n", landscape ? LANDSCAPE_HDR : PORTRAIT_HDR );
    }
    printf( "/bodyfontsize %g def\n", font_size );
    printf( "/lines %d def\n", linesperpage );
    printf( "/columns %d def\n", columnsperline );
    printf( "/date (%s %s) def\n", TodaysDate(), TodaysTime() );

    //  print PS header file
    while ( (c = getc(f)) != EOF )
        putchar(c);

    //  closing prologue
    printf( "%%%%EndProlog\n\n" );

    //  number of copies
    if ( copies_number > 1 )
        printf( "/#copies %d def\n", copies_number);

    printf( "/docsave save def\n" );
    printf( "startdoc\n" );
    if ( ! quiet_mode )
        fprintf( stderr, "Font size = %g points\n", font_size );
}


void    program_init( char *pg_name )
{
    char        *ptr ;

    tt_init( SHAREWARE_VERSION, 0, pgVERS_, pgTITL_, pgCOPR_ );
    logon(0);

    st_cpy( ps_header, pg_name, FS_LEN );
    ptr = strrchr(ps_header, '.');
    strcpy( ptr, ".HDR" );
    strupr( ps_header );
}


void    finish_off( void )
{
    printf( "\n%%%%Trailer\ncleanup\ndocsave restore end\n\x04" );
}


void    main( int argc, char *argv[] )
{
    register int narg ;
    char        *arg ;
    double      char_width, header_size, page_height, page_width ;

    program_init( argv[0] );

    //                                          process options
    arg = argv[narg = 1];
    while ( narg < argc && (arg[0] == '-' || arg[0] == '/') )
    {
        strlwr( arg );
        switch ( arg[1] )
        {
            case '?' :                          // help
                usage();
            case '#' :                          // n copies
                copies_number = 0;
                arg += 2;
                while ( *arg != EOS && isdigit(*arg) )
                    copies_number = copies_number*10 + (*arg++ - '0');
                if ( *arg != EOS || copies_number <= 0 )
                    usage();
                break;
            case '@' :                          // PS header file
                arg += 2;
                st_cpy( ps_header, arg, FS_LEN );
                strupr( ps_header );
                break;
            case '1' :                          // 132-column reports
                if ( arg[2] == '3' && arg[3] == '2' )
                {
                    reps_132 = TRUE ;
                    wide_pages = no_border = no_header = TRUE ;
                }
                else
                    usage();
                break;
            case 'b' :                          // print binary files
                if ( arg[2] != EOS )
                    usage();
                print_binaries = TRUE ;
                break;
            case 'f' :                          // change font size
                if ( arg[2] == EOS )
                {
                    folding = TRUE ;
                    break;
                }
                if ( (font_size = atof(&arg[2])) == 0.0 )
                    abort_job( "Wrong font size, use -f24.5 for example." );
                break;
            case 'h' :                          // help
                usage();
            case 'i' :                          // interpret control chars
                if ( arg[2] != EOS )
                    usage();
                interpret = TRUE;
                break;
            case 'l' :                          // left margin
                if ( arg[2] == EOS )
                    usage();
                if ( (left_margin = atoi(&arg[2])) <= 0 )
                    abort_job( "Wrong left margin size, use -l5 for a 5-char left margin." );
                break;
            case 'm' :                          // margins
                if ( arg[2] == EOS )
                    usage();
                if ( (margin = atof(&arg[2])) == 0.0 )
                    abort_job( "Wrong margin size, use -m1.2 for example." );
                break;
            case 'n' :                          // number file lines
                if ( arg[2] == EOS )
                {
                    numbering = TRUE;
                    break;
                }
                if ( arg[3] != EOS )
                    usage();
                switch (arg[2])
                {
                    case 'b' :                  // don't print binaries
                        print_binaries = FALSE;
                        break;
                    case 'f' :                  // cut lines too long
                        folding = FALSE;
                        break;
                    case 'h' :                  // don't print header
                        no_header = TRUE;
                        break;
                    case 'i' :                  // don't interpret ctrl chars
                        interpret = FALSE;
                        break;
                    case 'n' :                  // don't number lines
                        numbering = FALSE;
                        break;
                    case 'p' :                  // landscape format
                        landscape = TRUE;
                        break;
                    case 'r' :                  // don't restart sheet number
                        restart = FALSE;
                        break;
                    case 's' :                  // no surrounding border
                        no_border = TRUE;
                        break;
                    case 'v' :                  // only printable chars
                        only_printable = TRUE;
                        break;
                    case 'w' :                  // twin pages
                        wide_pages = FALSE;
                        break;
                    default :
                        usage();
                }
                break;
            case 'p' :                          // portrait format
                if ( arg[2] != EOS )
                    usage();
                if ( wide_pages )
                    abort_job( "Options -p and -w are incompatible." );
                landscape = FALSE;
                break;
            case 'q' :                          // quiet mode
                quiet_mode = TRUE ;
                break;
            case 'r' :                          // restart sheet number
                if ( arg[2] != EOS )
                    usage();
                restart = TRUE;
                break;
            case 't' :                          // set tab size
                if ( arg[2] == EOS || (column_width = atoi(arg+2)) <= 0 )
                    usage();
                break;
            case 'v' :                          // print control chars
                if ( arg[2] != EOS )
                    usage();
                only_printable = FALSE;
                break;
            case 'w' :                          // wide format
                if ( arg[2] != EOS )
                    usage();
                if ( ! landscape )
                    abort_job( "Options -p and -w are incompatible." );
                wide_pages = TRUE;
                break;
            default :
                usage();
        }
        arg = argv[++narg];
    }
    if ( arg != EOS && strcmp(arg, "?") == 0 )
        usage();

    if ( reps_132 )                             // 132-column reports...
    {
        font_size = 7.8 ;
        left_margin = 8 ;
    }

    twinpage = landscape && (! wide_pages) ;
    if ( font_size == 0.0 )
        font_size = landscape ? 6.8 : 9.0 ;
    page_height = (HEIGHT - margin) * PIXELS_INCH ;
    page_width = (WIDTH - margin) * PIXELS_INCH ;
    char_width = 0.6 * font_size ;
    if ( landscape )
    {
        header_size = no_header ? 0.0 : LANDSCAPE_HDR * PIXELS_INCH ;
        linesperpage = (int) (((page_width - header_size) / font_size) - 1.0);
        if ( wide_pages )
            columnsperline = (int) ((page_height / char_width) - 1.0);
        else
            columnsperline = (int) (((page_height / 2.0) / char_width) - 1.0);
    }
    else
    {
        header_size = no_header ? 0.0 : PORTRAIT_HDR * PIXELS_INCH ;
        linesperpage = (int) (((page_height - header_size) / font_size) - 1.0);
        columnsperline = (int) ((page_width / char_width) - 1.0);
    }
    if ( linesperpage <= 0 || columnsperline <= 0 )
        abort_job( "Font %g too big.", font_size );

    //                                  print files designated or standard input
    prefix_width = left_margin + (numbering ? 6 : 1) ;
    if ( narg >= argc )
    {
        print_file( "stdin" );
    }
    else
    {
        while ( narg < argc )
        {
            strupr( arg );
            if ( freopen(arg, "r", stdin) == NULPTR )
            {
                finish_off();
                abort_job( "Error opening %s", arg );
            }
            if ( ! quiet_mode )
                fprintf( stderr, "Printing %s\n", arg );
            print_file(arg);
            arg = argv[++narg];
        }
    }
    finish_off();
    if ( ! quiet_mode )
        fprintf( stderr, "\n%d page%s printed\n", sheets, ONE_OR_MORE(sheets) );
    exit( GOOD_EXIT );
}
